-- Example 2 - Database Corruption
-- Created by Steve Stedman  http://SteveStedman.com
-- Twitter @sqlEmt
-- LinkedIn http://linkedin.com/in/stevestedman
------------------------------------------------------------

USE [master]
GO
-- lets start by looking at what files the backup contains.
RESTORE FILELISTONLY
   FROM DISK = N'C:\DBBackups\CorruptionChallenge2_LatestBackup.bak' ;

RESTORE FILELISTONLY
   FROM DISK = N'C:\DBBackups\CorruptionChallenge2_TwoDaysAgoBackup.bak' ;
GO

RESTORE DATABASE [CorruptionChallenge2] 
   FROM DISK = N'C:\DBBackups\CorruptionChallenge2_LatestBackup.bak' 
   WITH FILE = 1,  
        MOVE N'CorruptionChallenge2' TO N'C:\SQL_DATA\CorruptionChallenge2.mdf',  
        MOVE N'CorruptionChallenge2_log' TO N'C:\SQL_DATA\CorruptionChallenge2_log.ldf',  
        RECOVERY, REPLACE;

GO
USE CorruptionChallenge2;
GO 

DBCC CheckDB(CorruptionChallenge2) WITH NO_INFOMSGS;

DBCC CheckTable(Revenue) WITH NO_INFOMSGS;

SELECT * 
  FROM Revenue;



-- so what does page 244 look like?
-- the first 96 bytes of a page are the page header...
DBCC TRACEON(3604);
DBCC PAGE('CorruptionChallenge2', 1, 244, 2);
--  ******** ****
--  020003ff ffff0001 f3000000 01001400 f5000000
-- byte
-- 0    HeaderVersion -- should be 01
-- 1    Type -- 01 for a datapage
-- 2    TypeFlagBits 
-- 3    Level - should be zero for data pages
-- 4-5  FlagBits  

-- m_slotCnt = 12  if it is not corrupt tells us there are 12 rows in this page.


DBCC PAGE('CorruptionChallenge2', 1, 245, 2);
--  01010400 00020001 f4000000 01001400 f6000000

DBCC PAGE('CorruptionChallenge2', 1, 243, 2);
--  01010400 00020001 f2000000 01001400 f4000000 


-- so at this point the page header looks corrupt in that one page.




-- Knowing we have one page corrupt, what data can we get from the existing database?

-- now what indexes do we have available?
SELECT ind.name as IndexName,
       col.name as ColumnName,
       ic.is_included_column
  FROM sys.indexes ind 
 INNER JOIN sys.index_columns ic ON  ind.object_id = ic.object_id and ind.index_id = ic.index_id 
 INNER JOIN sys.columns col ON ic.object_id = col.object_id and ic.column_id = col.column_id 
 WHERE ind.object_id = OBJECT_ID('Revenue') 
   AND ind.is_unique = 0 
   AND ind.is_unique_constraint = 0 
 ORDER BY ind.name, ind.index_id, ic.index_column_id;


SELECT id, DepartmentID, Revenue
  FROM Revenue
  WITH (INDEX (ncDeptIdYear) );
-- looks like there are 10057 rows in the index, and it doesn't appear to be corrupt


SELECT *
  FROM Revenue;
-- Returns 595 rows - up to ID #601


SELECT *
  FROM Revenue
ORDER BY id DESC;
-- returns 9450 rows back to id #615 


/*   Do the math
	Known that there are 10057 rows from nc index
	first query returned 595
	second query returned 9450

	10057
	- 595
	-9450
	=====
	   12 rows missing
*/

-- Lets find the ids of the missing rows.
SELECT id
  FROM Revenue
  WITH (INDEX (ncDeptIdYear) )
 WHERE id BETWEEN 602 AND 614
 ORDER BY id;

 -- missing values
 -- 602, 603, 604, 605, 606, 607, 608, 609, 611, 612, 613, 614


DBCC TRACEON(3604);
DBCC PAGE('CorruptionChallenge2', 1, 244, 1);

-- save off the output to compare later.




-- lets take a look at the previous backup and see what we have in that range.

RESTORE DATABASE [CorruptionChallenge2TwoDaysAgo] 
   FROM DISK = N'C:\DBBackups\CorruptionChallenge2_TwoDaysAgoBackup.bak' 
   WITH FILE = 1,  
        MOVE N'CorruptionChallenge2' TO N'C:\SQL_DATA\CorruptionChallenge2_oldBackup.mdf',  
        MOVE N'CorruptionChallenge2_log' TO N'C:\SQL_DATA\CorruptionChallenge2_oldBackup_log.ldf',  
        RECOVERY, REPLACE;

SELECT id
  FROM [CorruptionChallenge2TwoDaysAgo].[dbo].[Revenue]
 WHERE id BETWEEN 602 AND 614
 ORDER BY id;
 -- 13 rows


SELECT COUNT(1)
  FROM [CorruptionChallenge2TwoDaysAgo].[dbo].[Revenue]
-- only 10054 rows, not the original 10057 that we were looking for.

-- check just the ones we are looking for
SELECT *
  FROM [CorruptionChallenge2TwoDaysAgo].[dbo].[Revenue]
  WHERE ID IN (602, 603, 604, 605, 606, 607, 608, 609, 611, 612, 613, 614);
-- we have all 12 rows.  Do we know if anything was changed?  

DBCC TRACEON(3604);
DBCC PAGE('CorruptionChallenge2TwoDaysAgo', 1, 244, 1);
-- save the results....compare with a diff tool




-- now to save off the rows.
-- 

USE [CorruptionChallenge2]
GO

-- temp table to store some rows.
CREATE TABLE #RevenueSave(
	[id] [int] NOT NULL,
	[DepartmentID] [int] NULL,
	[Revenue] [int] NULL,
	[Year] [int] NULL,
	[Notes] [varchar](300) NULL,
) ;


-- save the rows before the corruption
INSERT INTO #RevenueSave
SELECT TOP 595 *
  FROM CorruptionChallenge2.dbo.Revenue 
 ORDER BY id ASC;

-- save the rows after the coruption
INSERT INTO #RevenueSave
SELECT TOP 9450 *
  FROM CorruptionChallenge2.dbo.Revenue 
 ORDER BY id DESC;

-- get the missing rows from the old database.
INSERT INTO #RevenueSave
SELECT *
  FROM [CorruptionChallenge2TwoDaysAgo].[dbo].[Revenue]
  WHERE ID IN (602, 603, 604, 605, 606, 607, 608, 609, 611, 612, 613, 614);


-- clean up the corruption by truncating the table
-- could use DBCC CheckTable(Revenue, REPAIR_ALLOW_DATA_LOSS)
TRUNCATE TABLE Revenue;

-- confirm corruption removed
DBCC CheckDB(CorruptionChallenge2) WITH NO_INFOMSGS;

-- check on our saved data
SELECT * 
  FROM #RevenueSave;
-- yep... all 10057 rows are still there



-- lets try putting it back.
INSERT INTO Revenue
SELECT * 
  FROM #RevenueSave;
-- whoops.



SET IDENTITY_INSERT Revenue ON;
INSERT INTO Revenue  
           ([id]
		   ,[DepartmentID]
           ,[Revenue]
           ,[Year]
           ,[Notes])
SELECT * 
  FROM #RevenueSave;
SET IDENTITY_INSERT Revenue OFF;
-- 10057 rows 


-- check the revenue table, confirm that it looks right.
SELECT * 
  FROM Revenue;
  
DBCC CheckDB(CorruptionChallenge2) WITH NO_INFOMSGS, ALL_ERRORMSGS;


-- cleanup our temp table
DROP TABLE #RevenueSave;

-- clean up our old backup database
ALTER DATABASE [CorruptionChallenge2TwoDaysAgo] SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
DROP DATABASE [CorruptionChallenge2TwoDaysAgo];


-- back up what we have as a new starting point.
BACKUP DATABASE CorruptionChallenge2
TO DISK = N'C:\DBBackups\CorruptionChallenge2_Fixed.bak' 
   WITH FORMAT,
   NAME = 'Full Backup of CorruptionChallenge2';



-- victory. No data loss.
